#ifndef __SINX_PLATFORM__
#define __SINX_PLATFORM__

//
// File: sinx_pltfm.h
// Customize this file to adapt SINX level 0 parser related classes to your
// particular platform
//

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <windows.h>
//#include <typeinfo>

/// @brief Cond is an expression. If it's false, it means logical error in
/// program described by msg, which is a string literal ("..."). User should
/// be notified, and program should be aborted.
/// Must have syntax of a statement.
/// In release config, it's allowed to replace this with empty definition.
#define SINX_ASSERT(cond,msg) do { if (!bool (cond)) {\
	printf ("Assertion failure: %s at %s, line %d\n",#cond,__FILE__,__LINE__);\
	DebugBreak ();\
	exit (1);\
} } while (0)

/// @brief SINX L0 parser classes are parametrized with PlatformL0 - a class
/// that must implement some platform binding interface. Here is a sample of
/// such class which contains all the necessary elements.
struct SINX0_Platform
{
	/// @brief Char type used by SINX parser(s)
	typedef char TChar;

	/// @brief Check whether TChar is a whitespace
	static bool IsWhitespace (TChar c) { return c>=0 && c<=32; }

	/// @brief TChar that is left angled bracket ('<')
	static const TChar KCharLAB = '<';

	/// @brief TChar that is right angled bracket ('>')
	static const TChar KCharRAB = '>';

	/// @brief TChar that is equality sign ('=')
	static const TChar KCharEqu = '=';

	/// @brief TChar that is dot ('.')
	static const TChar KCharDot = '.';

	/// @brief TChar that is colon (':')
	static const TChar KCharColon = ':';

	/// @brief TChar that is percent ('%')
	static const TChar KCharPercent = '%';

	/// @brief TChar that is null terminator
	static const TChar KCharNull = 0;

	/// @brief Allocate block capable of holding one instance of T (T is POD)
	/// plus immediately following array of aExtraBytes bytes and fill pDst
	/// with pointer to the block on success.
	/// On failure, throw exception, or abort, or whatever, just don't return.
	/// Semantics is somewhat strange, thanks to fucking GCC that doesn't
	/// understand explicit specification 'template member of a template param
	/// class'.
	/// (Note that all functions with L suffix are expected to behave this way.)
	template <class T>
	static void AllocL (T *&pDst,unsigned aExtraBytes)
	{
		pDst = (T *)new char[sizeof (T)+aExtraBytes];
	}

	/// @brief Deallocate block previously allocated by AllocL. Pointer is same
	/// as was set by AllocL.
	template <class T>
	static void Dealloc (T *ptr)
	{
		//printf ("Dealloc %s\n",typeid (T).name ());
		delete [] ptr;
	}

	/// @brief Implements a buffer of chars which holds data of SINX stream.
	/// Buffer is of fixed size and is assumed to not change while there are
	/// any SINX objects referencing to it.
	class Buffer
	{
		bool m_Allocated;
		int m_RefCount;
		const TChar *m_pData;
		unsigned m_DataLen;

	public:
		/// @brief Should return length of the buffer, in TChar-s
		unsigned Length () const { return m_DataLen; }

		/// @brief Buffer should have reference counter initially set to 0.
		/// AddRef should increase it by 1. Every SINX object that uses the
		/// Buffer keeps a reference to it. Also, AddRef is provided with
		/// aStartAt and aLen, hints that point actual portion referenced by
		/// this reference. It can be used to optimize buffering if the buffer
		/// implementation uses external storage.
		/// @param aStartAt Start of referenced area
		/// @param aLen Length of referenced area
		void AddRef (unsigned aStartAt,unsigned aLen) { m_RefCount++; }

		/// @brief Reduce reference counter by 1. It's up to Buffer to decide
		/// what to do if its reference counter drops to 0 as result of Unref.
		/// @param aStartAt Start of referenced area
		/// @param aLen Length of referenced area
		void Unref (unsigned aStartAt,unsigned aLen)
		{
			if (--m_RefCount<=0 && m_Allocated) Dealloc<Buffer> (this);
		}

		/// @brief Random read access to buffer. [0] is its 1st char,
		/// Length ()-1 is last char. Out-of-bound check is up to
		/// implementator's decision.
		TChar operator [] (unsigned aIdx) const
		{
			SINX_ASSERT (aIdx<m_DataLen,"Buffer::[]: index out of range");
			return m_pData[aIdx];
		}

		/// @brief Buffer should be able to be constructed over pre-existing
		/// buffer of chars. Lifetime of this buffer should be not lower than
		/// that of the Buffer, it's up to user to ensure it. It's also up to
		/// user to ensure that Buffer will not be outlived by any of SINX
		/// object that reference it (reference counting must be ignored if
		/// Buffer is constructed with this constructor).
		/// @param pData User-owned buffer of characters.
		Buffer (const TChar *pData,unsigned aLength):
			m_Allocated (false), m_RefCount (0),
			m_pData (pData), m_DataLen (aLength)
		{}

		/// @breif (Not part of SINX concept.) Buffer can be constructed to
		/// hold the data by itself. It's allocated dynamically, begins ref
		/// counting as first SINX object references it and automatically
		/// deallocates itself after no more references are left at it.
		/// Allocation is performed by SINX0_Pltfm::AllocL.
		static Buffer *AllocL (unsigned aLength)
		{
			Buffer *self;
			SINX0_Platform::AllocL (self,sizeof (TChar)*aLength);
			self->m_Allocated = true;
			self->m_RefCount = 0;
			self->m_pData = (const TChar *)(self+1);
			self->m_DataLen = aLength;
			return self;
		}

		/// @brief (Not part of SINX concept.) Return pointer to data of the
		/// buffer.
		const TChar *Data () const { return m_pData; }
	};
};

#endif // __SINX_PLATFORM__