#ifndef __SINX_READ_LEVEL0__
#define __SINX_READ_LEVEL0__

//
// Implementation of SINX level 0 parser related classes
//

#include "sinx_pltfm.h"

/// @brief An SINX symbol, plain or special.
template <class Platform>
class SINX0_Symbol;

/// @brief Read-only string obtained somehow from Platform::Buffer.
template <class Platform>
class SINX0_String
{
	friend class SINX0_Symbol<Platform>;

	typename Platform::Buffer *m_pBuffer;
	unsigned m_StartAt,m_Length;

	SINX0_String (typename Platform::Buffer *pBuffer,unsigned aStartAt,
		unsigned aLength):
		m_pBuffer (pBuffer), m_StartAt (aStartAt), m_Length (aLength)
	{ if (m_pBuffer!=NULL) m_pBuffer->AddRef (m_StartAt,m_Length); }

public:
	/// @brief Creates empty string (not assigned to any buffer).
	SINX0_String (): m_pBuffer (NULL), m_Length (0) {}

	/// @brief Creates string that contains all buffer data.
	/// @param pBuffer The buffer. Must not be NULL.
	SINX0_String (typename Platform::Buffer *pBuffer)
	{
		SINX_ASSERT (pBuffer!=NULL,"SINX0_String: buffer must not be null");
		m_pBuffer = pBuffer;
		m_StartAt = 0;
		m_Length = m_pBuffer->Length ();
		m_pBuffer->AddRef (m_StartAt,m_Length);
	}

	SINX0_String (const SINX0_String &src):
		m_pBuffer (src.m_pBuffer),
		m_StartAt (src.m_StartAt),
		m_Length (src.m_Length)
	{
		if (m_pBuffer!=NULL) m_pBuffer->AddRef (m_StartAt,m_Length);
	}

	/// @brief Assignment of one SINX0_String to another is allowed. Target
	/// SINX0_String will re-reference to same Platform::Buffer as src
	/// references.
	SINX0_String &operator = (const SINX0_String &src)
	{
		if (m_pBuffer!=NULL) m_pBuffer->Unref (m_StartAt,m_Length);
		m_pBuffer = src.m_pBuffer;
		m_StartAt = src.m_StartAt;
		m_Length = src.m_Length;
		if (m_pBuffer!=NULL) m_pBuffer->AddRef (m_StartAt,m_Length);
		return *this;
	}

	~SINX0_String ()
	{
		if (m_pBuffer!=NULL) m_pBuffer->Unref (m_StartAt,m_Length);
	}

	/// @brief Return the underlying buffer (can be NULL for empty strings)
	typename Platform::Buffer *Buffer () const { return m_pBuffer; }

	/// @brief Index of char in associated buffer at which the string starts
	unsigned StartAt () const { return m_StartAt; }

	/// @brief Length of the string, in Platform::TChar-s
	unsigned Length () const { return m_Length; }

	/// @brief Random read access to buffer. [0] is its 1st char,
	/// Length ()-1 is last char.
	typename Platform::TChar operator [] (unsigned aIdx) const
	{
		SINX_ASSERT (aIdx<m_Length,"SINX0_String::[]: index out of range");
		return (*m_pBuffer)[m_StartAt+aIdx];
	}

	/// @brief Create substring starting from character aStartAt and length of
	/// aLength (will be clamped if there are less characters after aStart).
	SINX0_String Substr (unsigned aStartAt,unsigned aLength) const
	{
		SINX_ASSERT (aStartAt<=m_Length,"SINX0_String::Substr: aStartAt is out of range");
		if (aStartAt+aLength>m_Length) aLength = m_Length-aStartAt;
		return SINX0_String<Platform> (m_pBuffer,aStartAt+m_StartAt,aLength);
	}

	bool operator == (const SINX0_String &aRhs) const
	{
		if (Length ()!=aRhs.Length ()) return false;
		else
		{
			unsigned i,aLen = Length ();
			for (i=0; i<aLen; i++) if ((*this)[i]!=aRhs[i]) return false;
			return true;
		}
	}

	bool operator != (const SINX0_String &aRhs) const
	{
		return !(*this==aRhs);
	}

	/// @brief SINX0_String can be used as index in std::map
	bool operator < (const SINX0_String &aRhs) const
	{
		unsigned i,aLenLhs = Length (),aLenRhs = aRhs.Length ();
		for (i=0; i<aLenLhs && i<aLenRhs; i++)
			if ((*this)[i]<aRhs[i]) return true;
		return aLenLhs<aLenRhs;
	}

	/// @brief Compare for equality with null-terminated C-style string
	bool operator == (const typename Platform::TChar *pCStr) const
	{
		unsigned i,aLen = Length ();
		for (i=0; i<aLen; i++, pCStr++)
		{
			if ((*this)[i]!=*pCStr ||
				// rhs string ended (by null term), we're not -> we're longer
				Platform::KCharNull==*pCStr)
					return false;
		}
		if (Platform::KCharNull!=*pCStr)
			return false; // pCStr is not over -> it is longer
		// unequality not detected -> equal
		return true;
	}

	bool operator != (const typename Platform::TChar *pCStr) const
	{
		return !(*this==pCStr);
	}
};

/// @brief SAX-like parser
template <class Platform>
class SINX0_SuxxParser;

/// @brief SINX element obtained by SAX-like parser. Can be either a run of
/// Platform::TChar-s, or opening of spec symbol, or closing of spec symbol,
/// or EOF marker (not actual part of the stream, just a marker that end of it
/// is reached).
template <class Platform>
class SINX0_SuxxElement
{
	friend class SINX0_SuxxParser<Platform>;

	SINX0_String<Platform> m_Data; // name of opening symbol or data of data run
	bool
		m_SpecAttr, // ESpecOpen only: symbol's name starts with ignorable '.'
		m_SpecTerm, // ESpecOpen only: symbol's name ends with ignorable '='
		m_SpecComp; // ESpecOpen only: symbol's name ends with ignorable ':'

public:
	/// @brief Possible types of SINX0_SuxxElement
	enum TType
	{
		/// @brief Run of data. Data () returns it.
		EDataRun,
		/// @brief Opening part of a special symbol. Data () returns special
		/// symbol name (w/o ignorable '.'-s, '='-s and ':'-s; presence of
		/// these traits can be determined via IsAttribute (), IsTerminal ()
		/// and IsCompound ())
		ESpecOpen,
		/// @brief Closing part of a special symbol (basically, '>')
		ESpecClose,
		/// @brief End of stream (the string over which SAX-like parser is
		/// constructed)
		EEof
	};

	/// @brief Returns entry type.
	TType Type () const
	{
		return m_Type;
	}

	/// @brief SINX0_SuxxElement is equivalent to false if its type is EEof,
	/// otherwise it's equivalent to true.
	operator bool () const
	{
		return m_Type!=EEof;
	}

	/// @brief Constructed by default, SINX0_SuxxElement has preset type of EEof.
	SINX0_SuxxElement ():
		m_SpecAttr (false),
		m_SpecTerm (false),
		m_SpecComp (false),
		m_Type (EEof)
		{}

	/// @brief Returns entry data. For EDataRun, contains the actual piece of
	/// data. For ESpecOpen, contains name of the spec symbol opened, with all
	/// ignorable '.', '=' and ':' trimmed. For other types, returns null
	/// SINX0_String.
	SINX0_String<Platform> Data () const { return m_Data; }

	/// @brief For ESpecOpen, returns true if symbol's name started with
	/// ignorable '.' (attribute metasyntax hint), otherwise false. For all
	/// other types, returns false.
	bool IsAttribute () const { return m_SpecAttr; }

	/// @brief For ESpecOpen, returns true if symbol's name ended with
	/// ignorable '=' (terminal metasyntax hint), otherwise false. For all
	/// other types, returns false.
	bool IsTerminal () const { return m_SpecTerm; }

	/// @brief For ESpecOpen, returns true if symbol's name ended with
	/// ignorable ':' (compound metasyntax hint), otherwise false. For all
	/// other types, returns false.
	bool IsCompound () const { return m_SpecComp; }

private:
	TType m_Type;
};

template <class Platform>
class SINX0_SuxxParser
{
	typedef SINX0_SuxxElement<Platform> TElement;
	SINX0_String<Platform> m_Src;
	unsigned m_Cursor;

public:
	/// @brief Construct SAX-like parser over string aSrc and reset cursor
	/// to beginning of it
	SINX0_SuxxParser (const SINX0_String<Platform> &aSrc):
		m_Src (aSrc), m_Cursor (0) {}

	/// @brief Returns current cursor position
	unsigned Cursor () const { return m_Cursor; }

	/// @brief Sets current cursor position to aCursor. aCursor must be either
	/// 0 (beginning of the source string) or a value obtained by Cursor (), or
	/// the parser may return incorrect (yet formally valid) result for
	/// GetNextElement. Cursor's value is offset from beginning of the source
	/// string in Platform::TChar-s, and therefore must not be set to value
	/// greater than string's length.
	void SetCursor (unsigned aCursor)
	{
		SINX_ASSERT (aCursor<=m_Src.Length (),
			"SINX0_SuxxParser::SetCursor: cursor value is out of range");
	}

	/// @brief Gets and returns next element of SINX string, and advances cursor
	/// to the next element. If cursor is at the end, returned element has type
	/// of SINX0_SuxxElement<Platform>::EEof.
	SINX0_SuxxElement<Platform> GetNextElement ()
	{
		TElement aResult;
		if (m_Cursor>=m_Src.Length ()) return aResult; // eof
		if (Platform::KCharLAB==m_Src[m_Cursor])
		{
			// cursor at opening '<': it's either spec open or beginning of
			// verbatim data run, name is expected - string of characters
			// except for '<', '>', '=', '%', whitespace or eof
			unsigned aNameStartsAt = ++m_Cursor,aNameLen = 0;
			for (;
				m_Cursor<m_Src.Length () &&
				!Platform::IsWhitespace (m_Src[m_Cursor]) &&
				Platform::KCharLAB!=m_Src[m_Cursor] &&
				Platform::KCharRAB!=m_Src[m_Cursor] &&
				Platform::KCharPercent!=m_Src[m_Cursor] &&
				Platform::KCharEqu!=m_Src[m_Cursor];
				m_Cursor++, aNameLen++) {}
			// remember name (or verbatim data delimiter)
			SINX0_String<Platform> aNameOrDelim =
				m_Src.Substr (aNameStartsAt,aNameLen);
			if (m_Cursor<m_Src.Length () && Platform::KCharPercent==m_Src[m_Cursor])
			{
				// next sign is '%' - it's a verbatim data run. Scan to the
				// ending '%' + delimiter or to eof, whatever gets first.
				// Start by skipping '%'.
				m_Cursor++;
				unsigned aDataStartAt = m_Cursor,aDataLen = 0;
			VERB_DELIM_REITERATE:
				for (;
					m_Cursor<m_Src.Length () &&
					Platform::KCharPercent!=m_Src[m_Cursor];
					m_Cursor++, aDataLen++) {}
				if (m_Cursor>=m_Src.Length ())
				{
					// we ran into eof - verbatim data run is unterminated,
					// return it as is
					aResult.m_Type = TElement::EDataRun;
					aResult.m_Data = m_Src.Substr (aDataStartAt,aDataLen);
					return aResult;
				}
				// we encountered a '%' - check if it's followed by matching
				// delimiter sequence
				unsigned aMatchedLen = 1,aCheckStartAt = m_Cursor;
				for (m_Cursor++; // skip '%'
					m_Cursor<m_Src.Length () &&
					aMatchedLen-1<aNameOrDelim.Length () &&
					m_Src[m_Cursor]==aNameOrDelim[aMatchedLen-1];
					m_Cursor++, aMatchedLen++) {}
				// check if we succeeded in match or not
				if (aMatchedLen-1>=aNameOrDelim.Length () &&
					// whole len matched, now expecting '>'
					m_Cursor<m_Src.Length () &&
					Platform::KCharRAB==m_Src[m_Cursor])
				{
					// found! it's a valid verbatim data end
					m_Cursor++; // skip '>'
					aResult.m_Type = TElement::EDataRun;
					// everything between '%'-s is the data
					aResult.m_Data = m_Src.Substr (aDataStartAt,
						aCheckStartAt-aDataStartAt);
					return aResult; // done with verbatim data run
				}
				// we didn't succeed in match, or it wasn't followed by '>'
				// therefore, consider all chars we just checked as part of
				// data, append them and proceed
				aDataLen += aMatchedLen;
				goto VERB_DELIM_REITERATE;
			}
			else
			{
				// next sign is not '%' - it was a spec sym name
				aResult.m_Type = TElement::ESpecOpen;
				if (m_Cursor<m_Src.Length ())
				{
					// it's not followed by eof
					if (Platform::KCharEqu==m_Src[m_Cursor])
					{
						// it's followed by '=' delimiter (terminal class hint)
						// skip the delimiter and mark result as terminal
						m_Cursor++;
						aResult.m_SpecTerm = true;
					}
					else if (Platform::IsWhitespace (m_Src[m_Cursor]))
					{
						// it's followed by whitespace
						// skip whitespaces
						for (; m_Cursor<m_Src.Length () &&
							Platform::IsWhitespace (m_Src[m_Cursor]);
							m_Cursor++) {}
					}
				}
				if (aNameOrDelim.Length ()>0 &&
					Platform::KCharDot==aNameOrDelim[0])
				{
					// name starts with '.' (attribute class hint)
					// strip the dot from name and mark result as attribute
					aNameOrDelim = aNameOrDelim.Substr (1,aNameOrDelim.Length ()-1);
					aResult.m_SpecAttr = true;
				}
				if (aNameOrDelim.Length ()>0 &&
					Platform::KCharColon==aNameOrDelim[aNameOrDelim.Length ()-1])
				{
					// name ends with ':' (compound class hint)
					// strip the colon from name and mark result as compound
					aNameOrDelim = aNameOrDelim.Substr (0,aNameOrDelim.Length ()-1);
					aResult.m_SpecComp = true;
				}
				aResult.m_Data = aNameOrDelim;
				return aResult;
			}
		}
		else if (Platform::KCharRAB==m_Src[m_Cursor])
		{
			// cursor at closing '>'
			aResult.m_Type = TElement::ESpecClose;
			m_Cursor++;
			return aResult;
		}
		else
		{
			// all other is data run: accumulate until eof, '<' or '>'
			unsigned aDataStartAt = m_Cursor,aLen = 0;
			for (; m_Cursor<m_Src.Length () &&
				Platform::KCharLAB!=m_Src[m_Cursor] &&
				Platform::KCharRAB!=m_Src[m_Cursor];
				m_Cursor++, aLen++) {}
			aResult.m_Type = TElement::EDataRun;
			aResult.m_Data = m_Src.Substr (aDataStartAt,aLen);
			return aResult;
		}
	}
};

/// @brief Class that provides iteration through a symbol's data according to
/// SINX level 0 concept
template <class Platform>
class SINX0_SymbolIterator;

template <class Platform>
class SINX0_Symbol
{
	friend class SINX0_SymbolIterator<Platform>;

	typedef SINX0_SuxxParser<Platform> SuxxParser;
	typedef SINX0_SuxxElement<Platform> SuxxElement;
	struct Entry
	{
		static const int KSpec = 1; // 0 == data run
		static const int KTerm = 2;
		static const int KAttr = 4;
		static const int KComp = 8;
		static const int KUnclosed = 16;
		static const int KFirst = 32;
		int m_Flags;
		union {
		Entry *m_pPrev; // re-use to store parent or prev sib during parse
		Entry *m_pNextSib;
		};
		Entry *m_p1stChild;
		unsigned m_StartAt,m_Length; // entry's data only, not name and brackets
		unsigned m_NameStartAt,m_NameLength; // entry's name (spec only)
	};

	struct Markup
	{
		int m_RefCount;
		typename Platform::Buffer *m_pBuffer;
		unsigned m_StartAt,m_Length,m_TotalEntries;
		// note that m_TotalEntries only specifies space allocated for parsing
		// and may be not equal to number of actually used entries (in case if
		// there are extra closes or unclosed symbols)
		Entry m_Entries[1];

		static Markup *AllocL (const SINX0_String<Platform> &aString,
			unsigned aExtraEntries)
		{
			Markup *pResult;
			Platform::AllocL (pResult,sizeof (Entry)*aExtraEntries);
			pResult->m_pBuffer = aString.Buffer ();
			pResult->m_StartAt = aString.StartAt ();
			pResult->m_Length = aString.Length ();
			if (pResult->m_pBuffer!=NULL)
				pResult->m_pBuffer->AddRef
					(aString.StartAt (),aString.Length ());
			pResult->m_RefCount = 1;
			return pResult;
		}
		void AddRef () { m_RefCount++; }
		void Unref ()
		{
			if (--m_RefCount<=0)
			{
				if (m_pBuffer) m_pBuffer->Unref (m_StartAt,m_Length);
				Platform::Dealloc (this);
			}
		}
	};

	Entry *FinalizeSubentries (Entry *pLastSubAdded,bool aExplicitClose,
		unsigned aCursorEnd)
	{
		// if last sub added is NULL (i. e., node to close has no data),
		// return NULL (we can't get anything more useful in this case anyway)
		if (pLastSubAdded==NULL) return NULL;

		// otherwise, return last completed node
		Entry *pEntry = pLastSubAdded,
			*pEntryBefore = pEntry->m_pPrev,
			*pEntryAfter = NULL;
		// correct cursor position from parser relative to buffer absolute
		aCursorEnd += m_pMarkup->m_Entries[0].m_StartAt;
		for (;;)
		{
			if (pEntry->m_Flags&Entry::KFirst) // reached 1st node, check parent
			{
				if (!aExplicitClose && pEntryBefore->m_Flags&Entry::KUnclosed)
				{
					// parent is unfinished - discard it and collapse to prev level
					pEntry->m_Flags &= ~Entry::KFirst;
					if (pEntryBefore->m_Flags&Entry::KFirst)
					{
						// also, inherit our first flag from parent (since we'll
						// mount ourselves on its place) and, if it's first in its
						// parent (our grandparent), change grandparent's first
						// child ptr to ourselves
						pEntry->m_Flags |= Entry::KFirst;
						pEntryBefore->m_pPrev->m_p1stChild = pEntry;
					}
					pEntryBefore = pEntryBefore->m_pPrev;
					continue;
				}
				else
				{
					// parent is finished (or explicit close specified) - stop here
					// and fix parent's data length
					pEntry->m_pNextSib = pEntryAfter;
					pEntryBefore->m_Flags &= ~Entry::KUnclosed;
					pEntryBefore->m_Length = aCursorEnd-pEntryBefore->m_StartAt;
					return pEntryBefore;
				}
			}
			// not 1st node: correct pointer and advance
			pEntry->m_pNextSib = pEntryAfter;
			pEntryAfter = pEntry;
			pEntry = pEntryBefore;
			pEntryBefore = pEntry->m_pPrev;
		}
	}

	Markup *m_pMarkup;
	union {
	Entry *m_pEntry;
	typename Platform::TChar m_AsChar;
	};

public:
	/// @brief Type of the symbol
	enum TType
	{
		/// @brief Plain character symbol, use AsChar() to get the value
		/// or compare to Platform::TChar
		EPlainChar,
		/// @brief Special symbol, use SpecialName() to get its name and
		/// SINX0_SymbolIterator to iterate thru its components
		ESpecial,
		/// @brief End-of-data marker. Compare Type() to this value to check
		/// if iteration through a ESpecial reached the end.
		EEof
	};
private:
	TType m_Type;

	SINX0_Symbol (Markup *pMarkup,Entry *pEntry):
		m_pMarkup (pMarkup), m_pEntry (pEntry), m_Type (ESpecial)
	{ if (m_pMarkup!=NULL) m_pMarkup->AddRef (); }

	SINX0_Symbol (typename Platform::TChar aChar):
		m_pMarkup (NULL), m_AsChar (aChar), m_Type (EPlainChar)
	{}

public:

	/// @brief Default SINX0_Symbol is of EEof type
	SINX0_Symbol (): m_pMarkup (NULL), m_pEntry (NULL), m_Type (EEof) {}

	~SINX0_Symbol () { if (m_pMarkup) m_pMarkup->Unref (); }

	SINX0_Symbol (const SINX0_Symbol &src):
		m_pMarkup (src.m_pMarkup), m_pEntry (src.m_pEntry), m_Type (src.m_Type)
	{ if (m_pMarkup!=NULL) m_pMarkup->AddRef (); }

	SINX0_Symbol &operator = (const SINX0_Symbol &src)
	{
		Markup *pPrevMarkup = m_pMarkup;
		m_pMarkup = src.m_pMarkup;
		m_pEntry = src.m_pEntry;
		m_Type = src.m_Type;
		if (m_pMarkup!=NULL) m_pMarkup->AddRef ();
		if (pPrevMarkup!=NULL) pPrevMarkup->Unref (); // unref previous data
		return *this;
	}

	/// @brief Parse aSrc as SINX level 0 compliant stream and set this
	/// SINX0_Symbol to 'virtual' spec symbol whose data is aSrc.
	/// Note: uses Platform::AllocL and can fail via it.
	SINX0_Symbol &operator = (const SINX0_String<Platform> &aSrc)
	{
		if (m_pMarkup) m_pMarkup->Unref (); // unref previous data
		m_pMarkup = NULL;
		m_pEntry = NULL;

		unsigned aEntryCount = 0;
		{
			SuxxParser aParser (aSrc);
			for (;;)
			{
				SuxxElement aElem = aParser.GetNextElement ();
				if (aElem.Type ()==SuxxElement::EEof)
					break;
				if (aElem.Type ()==SuxxElement::EDataRun ||
					aElem.Type ()==SuxxElement::ESpecOpen)
					aEntryCount++;
			}
		}

		// the only line where we can fail
		m_pMarkup = Markup::AllocL (aSrc,aEntryCount);
		m_pMarkup->m_TotalEntries = aEntryCount;

		unsigned aEntriesAdded = 1; // m_pMarkup->m_Entries[0] is the 1st,
		// it symbolizes the (virtual) root
		m_pMarkup->m_Entries[0].m_p1stChild = NULL;
		m_pMarkup->m_Entries[0].m_pNextSib = NULL;
		m_pMarkup->m_Entries[0].m_StartAt = aSrc.StartAt ();
		m_pMarkup->m_Entries[0].m_Length = aSrc.Length ();
		m_pMarkup->m_Entries[0].m_NameStartAt = aSrc.StartAt ();
		m_pMarkup->m_Entries[0].m_NameLength = 0;
		// 1st entry is by definition a special symbol, and it's not considered
		// as unclosed (for sake of finalization logic)
		m_pMarkup->m_Entries[0].m_Flags = Entry::KSpec;
		m_pEntry = &m_pMarkup->m_Entries[0]; // and we set this char to it
		Entry *pOvernode = &m_pMarkup->m_Entries[0],*pLastSubAdded = NULL;
		unsigned aLevel = 0;

		// scan thru components and subcomponents
		SuxxParser aParser (aSrc);
		for (;;)
		{
			SuxxElement aElem = aParser.GetNextElement ();
			switch (aElem.Type ())
			{
			case SuxxElement::EEof: goto MARKUP_PREPARED;
			case SuxxElement::ESpecOpen:
				{
					Entry &aNewSymb = m_pMarkup->m_Entries[aEntriesAdded++];
					aNewSymb.m_Flags = Entry::KSpec +
						(aElem.IsTerminal ()? Entry::KTerm : 0) +
						(aElem.IsCompound ()? Entry::KComp : 0) +
						(aElem.IsAttribute ()? Entry::KAttr : 0) +
						(pLastSubAdded? 0 : Entry::KFirst) +
						Entry::KUnclosed; // new spec char is yet unclosed
					if (pLastSubAdded!=NULL)
						// it's not first char in the subnode - link to
						// previous one
						aNewSymb.m_pPrev = pLastSubAdded;
					else
					{
						// it's first char in the subnode - link to subnode
						// (subnode is stored in pOvernode)
						pOvernode->m_p1stChild = &aNewSymb;
						aNewSymb.m_pPrev = pOvernode;
					}
					// enter newly opened char
					pOvernode = &aNewSymb;
					pLastSubAdded = NULL;
					aNewSymb.m_NameStartAt = aElem.Data ().StartAt ();
					aNewSymb.m_NameLength = aElem.Data ().Length ();
					aNewSymb.m_StartAt = aParser.Cursor ()+aSrc.StartAt ();
					aNewSymb.m_p1stChild = NULL;
					aLevel++;
				}
				break;
			case SuxxElement::ESpecClose:
				{
					// extra closes must be ignored
					if (aLevel>0)
					{
						// if close is not extra, finalize current char node
						// and return to prev level
						aLevel--;
						pLastSubAdded = FinalizeSubentries (pLastSubAdded,true,
							aParser.Cursor ()-1);
						if (pLastSubAdded==NULL)
						{
							// the symbol being closed was empty: finalize it
							// manually
							pLastSubAdded = pOvernode;
							pLastSubAdded->m_Flags &= ~Entry::KUnclosed;
							pLastSubAdded->m_Length =
								aParser.Cursor ()-1-pLastSubAdded->m_StartAt;
						}
					}
				}
				break;
			case SuxxElement::EDataRun:
				{
					// everything is simple in this case
					Entry &aNewSymb = m_pMarkup->m_Entries[aEntriesAdded++];
					aNewSymb.m_Flags = 0 +
						(pLastSubAdded? 0 : Entry::KFirst);
					if (pLastSubAdded!=NULL)
						// it's not first char in the subnode - link to
						// previous one
						aNewSymb.m_pPrev = pLastSubAdded;
					else
					{
						// it's first char in the subnode - link to subnode
						// (subnode is stored in pOvernode)
						pOvernode->m_p1stChild = &aNewSymb;
						aNewSymb.m_pPrev = pOvernode;
					}
					aNewSymb.m_StartAt = aElem.Data ().StartAt ();
					aNewSymb.m_Length = aElem.Data ().Length ();
					// name is dummy in DataRun
					aNewSymb.m_NameStartAt = aElem.Data ().StartAt ();
					aNewSymb.m_NameLength = 0;
					pLastSubAdded = &aNewSymb;
				}
				break;
			}
		}
		MARKUP_PREPARED:
		// reached end of stream, finalize all unclosed entries
		// special case when stream ends with sequence of unclosed opening
		// symbols - roll back over them until we run into something finished
		if (pLastSubAdded==NULL)
			for (pLastSubAdded=pOvernode;
				pLastSubAdded->m_Flags&Entry::KUnclosed;
				pLastSubAdded=pLastSubAdded->m_pPrev) {}
		if (pLastSubAdded!=&m_pMarkup->m_Entries[0])
			// if there is anything to finalize at all
			FinalizeSubentries (pLastSubAdded,false,aParser.Cursor ());

		m_Type = ESpecial;
		return *this;
	}

	/// @brief Return type of the symbol
	TType Type () const	{ return m_Type; }

	/// @brief Shortcut for Type ()==EEof
	bool IsEof () const { return m_Type==EEof; }

	/// @brief Shortcut for Type ()==ESpecial
	bool IsSpecial () const { return m_Type==ESpecial; }

	/// @brief Return true if the symbol is special symbol and has an attribute
	/// hint (dot started name)
	bool IsAttribute () const
	{
		return m_Type==ESpecial &&
			m_pEntry!=NULL &&
			(m_pEntry->m_Flags&Entry::KAttr);
	}

	/// @brief Return true if the symbol is special symbol and has a compound
	/// hint (colon ending name)
	bool IsCompound () const
	{
		return m_Type==ESpecial &&
			m_pEntry!=NULL &&
			(m_pEntry->m_Flags&Entry::KComp);
	}

	/// @brief Return true if the symbol is special symbol and has a terminal
	/// hint (equality sign name/data delimiter)
	bool IsTerminal () const
	{
		return m_Type==ESpecial &&
			m_pEntry!=NULL &&
			(m_pEntry->m_Flags&Entry::KTerm);
	}

	/// @brief If the symbol is EPlainChar, return the character it represents.
	/// For any other types, the value is undefined and has no meaning.
	typename Platform::TChar AsChar () const { return m_AsChar; }

	/// @brief Symbol is equivalent to false if it is Eof, otherwise it is
	/// equivalent to true
	operator bool () const { return !IsEof (); }

	/// @brief If the symbol is ESpecial, return its name. For other types,
	/// returns empty string.
	SINX0_String<Platform> Name () const
	{
		if (m_Type==ESpecial)
			return SINX0_String<Platform> (m_pMarkup->m_pBuffer,
				m_pEntry->m_NameStartAt,m_pEntry->m_NameLength);
		else return SINX0_String<Platform> ();
	}

	/// @brief If the symbol is ESpecial, return its data (as raw unparsed
	/// string). For other types, returns empty string.
	SINX0_String<Platform> Data () const
	{
		if (m_Type==ESpecial)
			return SINX0_String<Platform> (m_pMarkup->m_pBuffer,
				m_pEntry->m_StartAt,m_pEntry->m_Length);
		else return SINX0_String<Platform> ();
	}

	/// @brief Return symbol equal to Eof marker. You can compare it to symbol
	/// obtained from SINX0_SymbolIterator to determine if iterator reached the
	/// end.
	static SINX0_Symbol Eof () { return SINX0_Symbol<Platform> (); }

	/// @brief Compare to Platform::TChar. Returns true if type is EPlainChar
	/// and AsChar() value matches aRhs.
	bool operator == (typename Platform::TChar aRhs) const
	{ return m_Type==EPlainChar && m_AsChar==aRhs; }

	/// @brief Compare to Platform::TChar. Returns true if type is not
	/// EPlainChar or AsChar() value doesn't match aRhs.
	bool operator != (typename Platform::TChar aRhs) const
	{ return !(*this==aRhs); }

	/// @brief Compare to another SINX0_Symbol<Platform>. Return true if they
	/// match with types, name (if applicable) and data (if applicable).
	bool operator == (const SINX0_Symbol &aRhs) const
	{
		return (m_Type==EEof && aRhs.m_Type==EEof) ||
			(m_Type==EPlainChar && aRhs.m_Type==EPlainChar &&
				m_AsChar==aRhs.m_AsChar) ||
			(m_Type==ESpecial && aRhs.m_Type==ESpecial &&
				Name ()==aRhs.Name () &&
				Data ()==aRhs.Data ());
	}

	/// @brief Compare to another SINX0_Symbol<Platform> for inequality.
	bool operator != (const SINX0_Symbol &aRhs) const
	{ return !(*this==aRhs); }

	/// @brief Comparer to another SINX0_Symbol<Platform> for 'less than'.
	/// In order for SINX0_Symbol to be usable as map index (may be not
	/// reasonable option with long symbols, though).
	bool operator < (const SINX0_Symbol &aRhs) const
	{
		return int (m_Type)<int (aRhs.m_Type) ||
			(m_Type==EPlainChar && aRhs.m_Type==EPlainChar &&
			m_AsChar<aRhs.m_AsChar) ||
			(m_Type==ESpecial && aRhs.m_Type==ESpecial &&
			(
			Name ()<aRhs.Name () ||
			(Name ()==aRhs.Name () && Data ()<aRhs.Data ())
			));
	}

	typedef SINX0_SymbolIterator<Platform> Iterator;
};

template <class Platform>
class SINX0_SymbolIterator
{
public:
	typedef SINX0_Symbol<Platform> Symbol;

private:
	typename Symbol::Entry *m_pBaseEntry,*m_pCurrentEntry;
	Symbol m_Symbol;
	unsigned m_DataRunLeft;

	void PrepareSymbol ()
	{
		while (m_pCurrentEntry!=NULL &&
			!(m_pCurrentEntry->m_Flags&Symbol::Entry::KSpec))
		{
			m_DataRunLeft = m_pCurrentEntry->m_Length;
			if (m_DataRunLeft>0) break;
			// skip all data entries with zero length
			m_pCurrentEntry = m_pCurrentEntry->m_pNextSib;
		}
	}

public:
	SINX0_SymbolIterator ():
		m_pBaseEntry (NULL), m_pCurrentEntry (NULL)
	{}

	/// @brief Construct iterator to iterate over data of symbol sym, rewind it
	/// to beginning of the data
	SINX0_SymbolIterator (const Symbol &sym): m_Symbol (sym)
	{
		m_pBaseEntry = sym.Type ()==Symbol::ESpecial? sym.m_pEntry : NULL;
		Rewind ();
	}

	/// @brief Reset iterator to first entry of the symbol
	/// @return Reference to self (for chain expressions)
	SINX0_SymbolIterator &Rewind ()
	{
		m_pCurrentEntry = m_pBaseEntry!=NULL? m_pBaseEntry->m_p1stChild : NULL;
		PrepareSymbol ();
		return *this;
	}

	/// @brief Return symbol whose data this iterator is iterating through
	const Symbol &GetSymbol () const { return m_Symbol; }

	/// @brief Get next symbol from iterated symbol's data, don't advance
	/// cursor over it
	Symbol PeekNext () const
	{
		if (m_pCurrentEntry!=NULL)
		{
			if (m_pCurrentEntry->m_Flags&Symbol::Entry::KSpec)
				return Symbol (m_Symbol.m_pMarkup,m_pCurrentEntry);
			else
				return Symbol ((*m_Symbol.m_pMarkup->m_pBuffer)
					[m_pCurrentEntry->m_StartAt+m_pCurrentEntry->m_Length
					-m_DataRunLeft]);
		}
		else
		{
			return Symbol::Eof ();
		}
	}

	/// @brief Get next symbol from iterated symbol's data and advance cursor
	/// over it
	Symbol GetNext ()
	{
		if (m_pCurrentEntry==NULL) return Symbol::Eof ();
		else
		{
			Symbol aResult = PeekNext ();
			if ((m_pCurrentEntry->m_Flags&Symbol::Entry::KSpec) ||
				--m_DataRunLeft==0)
			{
				m_pCurrentEntry = m_pCurrentEntry->m_pNextSib;
				PrepareSymbol ();
			}
			return aResult;
		}
	}

	/// @brief Get next special symbol from iterated symbol's data, skipping
	// all plain char symbols before it, and advance cursor over it. If
	/// there are no more special symbols, return Symbol::Eof ().
	Symbol GetNextSpecial ()
	{
		while (m_pCurrentEntry!=NULL &&
			!(m_pCurrentEntry->m_Flags&Symbol::Entry::KSpec))
			m_pCurrentEntry = m_pCurrentEntry->m_pNextSib;
		PrepareSymbol ();
		return GetNext ();
	}

	/// @brief Set iterator to iterate over data of symbol sym, rewind it
	/// to beginning of the data
	SINX0_SymbolIterator &operator = (const Symbol &sym)
	{
		m_Symbol = sym;
		m_pBaseEntry = sym.Type ()==Symbol::ESpecial? sym.m_pEntry : NULL;
		return Rewind ();
	}

	/// @brief Return true if iterator is at end of iterated data
	bool IsAtEof () const { return m_pCurrentEntry==NULL; }

	// assignment of SINX0_SymbolIterator to another SINX0_SymbolIterator is
	// not explicitly implemented, rely on compiler's default implementation
};

#endif // __SINX_READ_LEVEL0__
