#ifndef __SINX_READ_LEVEL1__
#define __SINX_READ_LEVEL1__

//
// Implementation of SINX level 1 parser related classes. Uses STL string, deque
// and map.
// Assumes platform is SINX0_Platform from default sdsl_pltfm.h. If you use
// different one, replace typedef of SINX1_Platform.
//

#include "sinx_read_l0.h"
#include <string.h>
#include <string>
#include <map>
#include <deque>

/// @brief Change if you use platform class different from default
/// SINX0_Platform.
typedef SINX0_Platform SINX1_Platform;

/// @brief A number of helper functions and typedefs. Some of them can be
/// redefined if their default implementations don't suit your platform, some
/// don't have to.
struct SINX1_Util
{
	/// @brief Helper function, check if characrer is a hex digit (case
	/// insensitive), returns -1 if no or value of the digit otherwise. Redefine
	/// if default implementation doesn't suit your platform.
	static int AsHexDigit (SINX1_Platform::TChar aChar)
	{
		return (aChar>='0' && aChar<='9')? int (aChar-'0') :
			((aChar>='A' && aChar<='F')? int (aChar-'A')+10 :
			((aChar>='a' && aChar<='f')? int (aChar-'a')+10
			: -1));
	}

	/// @brief Helper function, returns true if characrer is a 'x' letter (case
	/// insensitive) or false otherwise. Redefine if default implementation doesn't
	/// suit your platform.
	static bool IsX (SINX1_Platform::TChar aChar)
	{ return aChar=='x' || aChar=='X'; }

	/// @brief Helper function, returns true if characrer is a 'l' letter (case
	/// insensitive) or false otherwise. Redefine if default implementation doesn't
	/// suit your platform.
	static bool IsL (SINX1_Platform::TChar aChar)
	{ return aChar=='l' || aChar=='L'; }

	/// @brief Helper function, returns true if characrer is a 'g' letter (case
	/// insensitive) or false otherwise. Redefine if default implementation doesn't
	/// suit your platform.
	static bool IsG (SINX1_Platform::TChar aChar)
	{ return aChar=='g' || aChar=='G'; }

	/// @brief Helper function, returns true if characrer is a 't' letter (case
	/// insensitive) or false otherwise. Redefine if default implementation doesn't
	/// suit your platform.
	static bool IsT (SINX1_Platform::TChar aChar)
	{ return aChar=='t' || aChar=='T'; }

	/// @brief Code of minus char ('-') in the platform used. Redefine if it's
	/// different for you.
	static const SINX1_Platform::TChar KCharMinus = '-';

	/// @brief Code of hash char ('#') in the platform used. Redefine if it's
	/// different for you.
	static const SINX1_Platform::TChar KCharHash = '#';

	/// @brief Length of C-style null-terminated string pString. Redefine if
	/// it's different for you.
	static unsigned StrLen (const SINX1_Platform::TChar *pString)
	{ return (unsigned)strlen (pString); }

	//
	// Implementation part you likely won't have to customize
	//

	typedef std::basic_string<SINX1_Platform::TChar> String;

	/// @brief Helper function, checks if special symbol's name (given in aName)
	/// corresponds to a character ASCII/UCS16 code specification according to
	/// SINX level 1 (<#DecimalNumber> or <#0xHexNumber>). If yes, returns that
	/// code, otherwise returns -1.
	static int CheckCharCode (
		const SINX0_String<SINX1_Platform> &aName)
	{
		unsigned i;
		if (aName.Length ()>1 && aName[0]==SINX1_Util::KCharHash)
		{
			int aBase;
			if (aName.Length ()>=3 &&
				SINX1_Util::AsHexDigit (aName[1])==0 &&
				SINX1_Util::IsX (aName[2]))
			{
				// it starts with #0x... - assume hex code
				aBase = 16;
				i = 3;
			}
			else
			{
				// otherwise, assume dec code
				aBase = 10;
				i = 1;
			}
			int aResult = 0,aDigits = 0;
			for (; i<aName.Length (); i++, aDigits++)
			{
				int aDigit = SINX1_Util::AsHexDigit (aName[i]);
				if (aDigit<0 || aDigit>=aBase)
					// incorrect digit for a chosen radix - it's not a char code
					return -1;
				aResult = aResult*aBase+aDigit;
			}
			// consider it success if whole name is valid number of at least 1 digit
			return aDigits>0? aResult : -1;
		}
		else if (aName.Length ()==2 &&
			SINX1_Util::IsL (aName[0]) &&
			SINX1_Util::IsT (aName[1]))
			return '<';
		else if (aName.Length ()==2 &&
			SINX1_Util::IsG (aName[0]) &&
			SINX1_Util::IsT (aName[1]))
			return '>';
		else return -1; // not starting from '#' or too short
	}

	/// @brief Helper function, checks if special symbol's name (given in aName)
	/// corresponds to ignorable comment
	static bool IsComment (const SINX0_String<SINX1_Platform> &aName)
	{
		return aName.Length ()==2 &&
			KCharMinus==aName[0] &&
			KCharMinus==aName[1];
	}

	/// @brief Helper function, converts platform level 0 layer const string to
	/// modifiable string. May throw memory exception from STL internals.
	static String ToStringL (const SINX0_String<SINX1_Platform> &aSrc)
	{
		String aResult;
		aResult.resize (aSrc.Length ());
		unsigned i;
		for (i=0; i<aSrc.Length (); i++) aResult[i] = aSrc[i];
		return aResult;
	}

	/// @brief Return character by its ASCII/UCS16 code.
	static SINX1_Platform::TChar CharFromCode (int aCode)
	{ return SINX1_Platform::TChar (aCode); }

	/// @brief Return character's ASCII/UCS16 code.
	static int CodeFromChar (SINX1_Platform::TChar aChar)
	{ return int (aChar); }

	/// @brief Convert a C-style (null-terminated) string of system char to
	/// std::basic_string<SINX1_Platform::TChar>. Allowed to fail with
	/// exception. Doesn't need to control validity of pLit pointer.
	static String StringFromU8LitL (const char *pLit)
	{
		String aResult;
		unsigned i,n = (unsigned)strlen (pLit);
		aResult.resize (n);
		for (i=0; i<n; i++) aResult[i] = pLit[i];
		return aResult;
	}
};

/// @brief Representation of SINX special symbol as a compound symbol according
/// to SINX level 1. Comments (<-- ... -->) and character spec symbols
/// (<#DEC>, <#0xHEX>) are ignored, other characters are grouped by name
class SINX1_Symbol
{
public:
	typedef SINX0_Symbol<SINX1_Platform> L0_Symbol;

private:
	L0_Symbol m_Symbol;
	mutable bool m_ByNameMapBuilt;
	mutable std::map<SINX1_Util::String,L0_Symbol> m_ByNameMap;

	void BuildByNameMapL () const
	{
		m_ByNameMap.clear ();
		SINX0_SymbolIterator<SINX1_Platform> aIt = m_Symbol;
		for (;;)
		{
			L0_Symbol aSym = aIt.GetNext ();
			if (aSym.IsEof ()) break; // end of stream - finish
			if (!aSym.IsSpecial ()) continue; // plain char - skip
			if (SINX1_Util::IsComment (aSym.Name ()))
				// comment type symbol - skip
				continue;
			if (SINX1_Util::CheckCharCode(aSym.Name ())!=-1)
				// code-specified plain character type symbol - skip
				continue;
			// remember symbol. If there are more than one with same name,
			// later one overrides earlier ones
			m_ByNameMap[SINX1_Util::ToStringL (aSym.Name ())] = aSym;
		}
		m_ByNameMapBuilt = true;
	}

	static SINX1_Util::String ExtractStringL (L0_Symbol::Iterator &it)
	{
		// read string from current iterator, decoding character code entities
		// and skipping comments according to SINX level 1. Stop at 1st special
		// symbol that doesn't resolve to plain char or comment (or at eof),
		// place iterator at it.
		SINX1_Util::String aResult;
		int i;
		unsigned aStrLen = 0,u = 0;
		for (i=0; i<2; i++)
		{
			// parse in 2 iterations: 1st - count number of chars, 2nd - fill
			// the string
			if (i==1) aResult.resize (aStrLen);
			L0_Symbol::Iterator aCurIt = it;
			for (;;aCurIt.GetNext ())
			{
				L0_Symbol aSymbol = aCurIt.PeekNext ();
				if (aSymbol.IsEof ()) break;
				if (aSymbol.IsSpecial ())
				{
					// skip comments
					if (SINX1_Util::IsComment (aSymbol.Name ())) continue;
					int aCharCode = SINX1_Util::CheckCharCode (aSymbol.Name ());
					if (aCharCode==-1)
						// stop on 1st non-char symbol
						break;
					if (i==1) aResult[u++] =
						SINX1_Util::CharFromCode (aCharCode);
					else aStrLen++;
				}
				else
				{
					if (i==1) aResult[u++] = aSymbol.AsChar ();
					else aStrLen++;
				}
			}
			if (i==1) it = aCurIt;
		}
		return aResult;
	}

public:
	SINX1_Symbol (): m_ByNameMapBuilt (false) {}

	SINX1_Symbol (const L0_Symbol &src):
		m_Symbol (src),
		m_ByNameMapBuilt (false)
	{}
	SINX1_Symbol &operator = (const L0_Symbol &src)
	{
		m_Symbol = src;
		m_ByNameMap.clear ();
		m_ByNameMapBuilt = false;
		return *this;
	}

	SINX1_Symbol (const SINX1_Symbol &src):
		m_Symbol (src.m_Symbol),
		m_ByNameMapBuilt (false)
	{}
	SINX1_Symbol &operator = (const SINX1_Symbol &src)
	{
		m_Symbol = src.m_Symbol;
		m_ByNameMap.clear ();
		m_ByNameMapBuilt = false;
		return *this;
	}

	/// @brief This constructor may fail via PlatformL0::AllocL on OOM
	SINX1_Symbol (const SINX0_String<SINX1_Platform> &src):
		m_ByNameMapBuilt (false)
	{ m_Symbol = src; }
	/// @brief This assignment may fail via PlatformL0::AllocL on OOM
	SINX1_Symbol &operator = (const SINX0_String<SINX1_Platform> &src)
	{
		m_Symbol = src;
		m_ByNameMap.clear ();
		m_ByNameMapBuilt = false;
		return *this;
	}

	/// @brief Returns symbol's underlying level 0 abstraction
	L0_Symbol L0 () const { return m_Symbol; }

	/// @brief Shortcut for L0 ().IsEof ()
	bool IsEof () const { return m_Symbol.IsEof (); }

	/// @brief Symbol is equivalent to false if it is Eof, otherwise it is
	/// equivalent to true
	operator bool () const { return !m_Symbol.IsEof (); }

	/// @brief Shortcut for L0 ().IsSpecial ()
	bool IsSpecial () const { return m_Symbol.IsSpecial (); }

	/// @brief Shortcut for L0 ().IsTerminal ()
	bool IsTerminal () const { return m_Symbol.IsTerminal (); }

	/// @brief Shortcut for L0 ().IsAttribute ()
	bool IsAttribute () const { return m_Symbol.IsAttribute (); }

	/// @brief Get special subsymbol by name. May throw memory exception from
	/// STL internals.
	SINX1_Symbol operator [] (const SINX1_Util::String &aName) const
	{
		if (!m_ByNameMapBuilt) BuildByNameMapL ();
		std::map<SINX1_Util::String,L0_Symbol>::iterator it =
			m_ByNameMap.find (aName);
		if (it==m_ByNameMap.end ()) return L0_Symbol::Eof ();
		else return it->second;
	}

	/// @brief Version of operator [] to use with string literal
	SINX1_Symbol operator [] (const SINX1_Platform::TChar *pName) const
	{ return (*this)[SINX1_Util::String (pName)]; }

	/// @brief Get all special subsymbols, their order is same as one in the
	/// stream. May throw memory exception from STL internals.
	/// If aIncludeAll is true, all special symbols are included. Otherwise,
	/// comments and character codes (which formally are special symbols too)
	/// are not included.
	std::deque<SINX1_Symbol> GetAllSpecialL (bool aIncludeAll=false) const
	{
		std::deque<SINX1_Symbol> aResult;
		L0_Symbol::Iterator it = m_Symbol;
		L0_Symbol aSubsym;
		for (aSubsym=it.GetNextSpecial (); !aSubsym.IsEof ();
			aSubsym=it.GetNextSpecial ())
			if (aIncludeAll || (SINX1_Util::CheckCharCode (aSubsym.Data ())==-1
				&& !SINX1_Util::IsComment (aSubsym.Data ())))
				aResult.push_back (aSubsym);
		return aResult;
	}

	/// @brief Get all special subsymbols with given name, their order is same
	/// as one in the stream. May throw memory exception from STL internals.
	std::deque<SINX1_Symbol> GetAllSpecialL (const SINX1_Util::String &aName) const
	{
		std::deque<SINX1_Symbol> aResult;
		L0_Symbol::Iterator it = m_Symbol;
		L0_Symbol aSubsym;
		for (aSubsym=it.GetNextSpecial (); !aSubsym.IsEof ();
			aSubsym=it.GetNextSpecial ())
			if (aSubsym.Name ()==aName.c_str ())
				aResult.push_back (aSubsym);
		return aResult;
	}

	/// @brief Version of GetAllSpecialL to use with string literal.
	/// May throw memory exception from STL internals.
	std::deque<SINX1_Symbol> GetAllSpecialL (const SINX1_Platform::TChar *pName) const
	{ return GetAllSpecialL (SINX1_Util::String (pName)); }

	/// @brief Get symbol's name as SINX1_Util::String. If name is not
	/// applicable, returns empty string.
	/// May throw memory exception from STL internals.
	SINX1_Util::String NameL () const
	{ return SINX1_Util::ToStringL (m_Symbol.Name ()); }

	/// @brief Get symbol's raw data (unparsed SINX string) as
	/// SINX1_Util::String. If data is not applicable, return empty string.
	SINX1_Util::String DataL () const
	{ return SINX1_Util::ToStringL (m_Symbol.Data ()); }

	/// @brief Interpret symbol's data as string, decoding character code
	/// special symbols and ignoring comments accorting to SINX level 1. If
	/// there are special symbols different from comments and character codes,
	/// data is considered to be not interpretable as string, and aDefault is
	/// returned
	/// May throw memory exception from STL internals.
	SINX1_Util::String DataAsStringL (const SINX1_Util::String &aDefault=
		SINX1_Util::String ()) const
	{
		L0_Symbol::Iterator it = m_Symbol;
		SINX1_Util::String aResult = ExtractStringL (it);
		return it.IsAtEof ()? aResult : aDefault;
	}

	/// @brief Interpret symbol's data as string, decoding character code
	/// special symbols and ignoring comments accorting to SINX level 1, and
	/// then parse it as integer number (0xHex or Dec, with or without '-').
	/// On failure, return value provided int aDefault.
	/// May throw memory exception from STL internals.
	template <class I>
	I DataAsCustomIntL (I aDefault) const
	{
		L0_Symbol::Iterator it = m_Symbol;
		SINX1_Util::String str = ExtractStringL (it);
		if (it.PeekNext ())
			// there are more symbols - data can't be interpreted as integer
			return aDefault;

		I aBase;
		bool aMinus = false;
		unsigned i = 0;
		if (str.size ()>=1 && str[i]==SINX1_Util::KCharMinus)
		{
			// starts with minus
			aMinus = true;
			i++;
		}
		if (str.size ()+i>=2 &&
			SINX1_Util::AsHexDigit (str[i])==0 &&
			SINX1_Util::IsX (str[i+1]))
		{
			// it starts with 0x... - assume hex number
			aBase = I (16);
			i += 2;
		}
		else
			// otherwise, assume dec number
			aBase = I (10);
		I aResult = 0;
		int aDigits = 0;
		for (; i<str.size (); i++, aDigits++)
		{
			int aDigit = SINX1_Util::AsHexDigit (str[i]);
			if (aDigit<I (0) || aDigit>=aBase)
				// incorrect digit for a chosen radix - it's not a char code
				return aDefault;
			aResult = aResult*aBase+aDigit;
		}
		// consider it success if whole string is valid number of at least 1
		// digit
		return aDigits>0? (aMinus? -aResult : aResult) : aDefault;
	}

	/// @brief Interpret symbol's data as string, decoding character code
	/// special symbols and ignoring comments accorting to SINX level 1, and
	/// then parse it as integer number (0xHex or Dec, with or without '-').
	/// On failure, return value provided int aDefault.
	/// May throw memory exception from STL internals.
	int DataAsIntL (int aDefault) const
	{
		return DataAsCustomIntL<int> (aDefault);
	}

	/// @brief Return symbol that is located in this one's tree under a
	/// specified path. Path is an SINX string with structure that corresponds
	/// to symbol location. Example:
	/// &lt;char1 &lt;char2 &lt;char3&gt;&gt;&lt;.index=2&gt;&gt; means: find
	/// first spec symbol with name char1 in this, open it and find 3rd (0
	/// based index 2) symbol with name char2 in it, open it and find 1st symbol
	/// with name char3 in it, and that char3 will be the result.
	/// If the result is not found at given location, return Eof type symbol.
	/// Function may fail in an exception or allocation.
	SINX1_Symbol OpenPathL (const SINX1_Util::String &aPath)
	{
		// this function is also an example of SINX practical usage. See how
		// simple it is.
		SINX1_Platform::Buffer b (aPath.c_str (),(unsigned)aPath.size ());
		SINX1_Symbol aPathSym (&b),aResult = *this;
		SINX1_Util::String index = SINX1_Util::StringFromU8LitL ("index");
		for (;;)
		{
			SINX1_Symbol aPathSubsym =
				L0_Symbol::Iterator (aPathSym.L0 ()).GetNextSpecial ();
			if (aPathSubsym.IsEof ())
				// no more nested specs -> last open symbol is the one we seek
				return aResult;
			// otherwise, try to find it
			SINX1_Util::String aNextName = aPathSubsym.NameL ();
			std::deque<SINX1_Symbol> aSymbols =
				aResult.GetAllSpecialL (aNextName);
			unsigned aIdx = aPathSym[index].DataAsIntL (0);
			if (aIdx>aSymbols.size ())
				// index out of range - path becomes invalid at this point,
				// result is not found
				return L0_Symbol::Eof ();
			// found, proceed next iteration with what we have found
			aResult = aSymbols[aIdx];
			aPathSym = aPathSubsym;
		}
	}

	/// @brief Version of OpenPathL for argument given as string literal
	SINX1_Symbol OpenPathL (const SINX1_Platform::TChar *pPath)
	{ return OpenPathL (SINX1_Util::String (pPath)); }

	/// @brief Represents an entry in array of content - a string run or a spec
	/// symbol
	class Fragment
	{
		friend class SINX1_Symbol;

		L0_Symbol m_Symbol;
		SINX1_Util::String m_String;
		bool m_IsSymbol;

		Fragment (const L0_Symbol &src):
			m_Symbol (src), m_IsSymbol (true) {}
		Fragment (const SINX1_Util::String &src):
			m_String (src), m_IsSymbol (false) {}

	public:
		Fragment (): m_IsSymbol (false) {}
		bool IsSpecial () const { return m_IsSymbol; }
		bool IsString () const { return !m_IsSymbol; }
		SINX1_Symbol AsSpecial () const { return m_Symbol; }
		SINX1_Util::String AsStringL () const { return m_String; }
	};

	/// @brief Get all data (plain strings and special symbols) as array of
	/// Fragment-s. Character codes are interpreted as plain chars and are
	/// included in strings, comments are ignored.
	std::deque<Fragment> GetAllL () const
	{
		std::deque<Fragment> aResult;
		L0_Symbol::Iterator it = m_Symbol;
		for (; !it.IsAtEof ();)
		{
			SINX1_Util::String aString = ExtractStringL (it);
			if (aString.size ()>0) aResult.push_back (Fragment (aString));
			if (!it.IsAtEof ()) aResult.push_back (Fragment (it.GetNext ()));
		}
		return aResult;
	}
};

#endif // __SINX_READ_LEVEL0__
