#ifndef __SINX_WRITE_LEVEL1__
#define __SINX_WRITE_LEVEL1__

//
// Implementation of SINX writer related classes. Uses STL string and
// SINX1_Platform from level 1 reader.
//

#include "sinx_read_l1.h"

class SINX_WriteDocumentBase;

/// @brief Common part for SINX writing class. Symbolizes a special symbol
/// (which is treated as 'document node' in this case) open for write.
/// Everything you write to a node must be written with WriteString methods of
/// that node's object. As soon as object exists and is not explicitly closed,
/// the node is considered open, and it can not be written to if there are
/// open nodes inside it. (I. e., you can only WriteString to node that is most
/// recently open of currently existing nodes.)
class SINX_WriteNode
{
	friend class SINX_WriteSpecialOpen;
	friend class SINX_WriteDocumentBase;

	void DoWriteSpecial (const SINX1_Platform::TChar *pSpecName,
		const SINX1_Platform::TChar *pData,unsigned aCharCount,
		int aHints);
	void WriteIndent (int aSpaceCount);
	void DoWriteInt (unsigned aVal,bool aHex,bool aMinus);
	template <class U,int MAX_CHARS>
	void DoPrepareCustomInt (U aVal,bool aHex,bool aMinus,
		SINX1_Platform::TChar *&pString,int &aCount)
	{
		// fill pString in reverse order and shift the pointer
		int aDigits[16] =
		{ '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
			'A', 'B', 'C', 'D', 'E', 'F' };
		U aBase = aHex? U (16) : U (10);
		// construct string in reverse order
		int i = 0;
		for (; aVal>0; aCount++, i++, aVal/=aBase)
		{
			SINX_ASSERT (i<MAX_CHARS-3,"DoPrepareCustomInt: number too long");
			*(pString--) = SINX1_Util::CharFromCode (aDigits[aVal%aBase]);
		}
		if (i==0)
		{
			// case of 0
			*(pString--) = SINX1_Util::CharFromCode (aDigits[0]);
			aCount++;
		}
		if (aHex)
		{
			*(pString--) = SINX1_Util::CharFromCode ('x');
			*(pString--) = SINX1_Util::CharFromCode ('0');
			aCount += 2;
		}
		if (aMinus)
		{
			*(pString--) = SINX1_Util::KCharMinus;
			aCount++;
		}
	}

	void DoPrepareInt (unsigned aVal,bool aHex,bool aMinus,
		SINX1_Platform::TChar *&pString,int &aCount)
	{
		DoPrepareCustomInt<unsigned,64> (aVal,aHex,aMinus,pString,aCount);
	}

protected:
	SINX_WriteDocumentBase *m_pBase;
	SINX_WriteNode *m_pPrev;
	bool
		m_Empty, // if true, next WriteString/Special will emit a leading space
		m_NewlineUsed, // set to true whenever special with KHintNewline is out,
		// if true, closing '>' of the symbol will be emitted with newline and
		// indent as well
		m_Compound, // if true, each non-HintNewline special will be written
		// after a space, unless m_WasString is not true
		m_WasString;

	SINX_WriteNode ():
		m_pBase (NULL), m_pPrev (NULL),
		m_Empty (true), m_NewlineUsed (false),
		m_Compound (false), m_WasString (false) {}
	void PushOpen (SINX_WriteNode *);
	void PopClose ();
	virtual ~SINX_WriteNode () { PopClose (); }

public:
	/// @brief Use this function to append a string of data into currently open
	/// node. Will fail assertion if there the node contains unclosed subnodes.
	void WriteString (const SINX1_Platform::TChar *pData,unsigned aCharCount);

	/// @brief Version of WriteString that takes C-style null-terminated string
	/// as an argument
	void WriteString (const SINX1_Platform::TChar *pData)
	{ WriteString (pData,SINX1_Util::StrLen (pData)); }

	/// @brief Version of WriteString that takes SINX1_Util::String as an
	/// argument
	void WriteString (const SINX1_Util::String &aData)
	{ WriteString (aData.c_str (),(unsigned)aData.size ()); }

	/// @brief Writes string that represents an integer
	void WriteInt (int aInt)
	{
		if (aInt>=0) DoWriteInt ((unsigned)aInt,false,false);
		else DoWriteInt ((unsigned)-aInt,false,true);
	}

	/// @brief Writes string that represents an unsigned integer (in hex format)
	void WriteUnsigned (int aUnsigned)
	{ DoWriteInt (aUnsigned,true,false); }

	/// @brief Flag for WriteSpecial, denotes that symbol should be emitted
	/// with attribute hint
	static const int KHintAttribute = 1;

	/// @brief Flag for WriteSpecial, denotes that symbol should be emitted
	/// with compound hint
	static const int KHintCompound = 2;

	/// @brief Flag for WriteSpecial, denotes that symbol should be emitted
	/// with terminal hint
	static const int KHintTerminal = 4;

	/// @brief Flag for WriteSpecial, denotes that symbol should be emitted
	/// with newline and indent. Use inside symbols opened with KHintCompound
	/// to arrange a pretty formatting.
	static const int KHintNewline = 8;

	/// @brief Write a special symbol with given name and string data, both
	/// given with C-style null-terminated strings pSpecName and pData
	/// respectively. When writing, consider hint flags aHints (KHint...
	/// combined with | or +). The symbol written with this function is closed
	/// and completed. String is properly SINX-escaped, so it isn't possible to
	/// write a data with embedded special symbols with this function (and
	/// aHints is therefore always forced to have KHintTerminal for non-empty
	/// data, even if you don't specify it explicitly). If you want to write a
	/// symbol with complex data, consider using SINX_WriteSpecialOpen class.
	/// Note: is pSpecName contains characters illegal for name, all or part of
	/// it may be truncated and transfer to data.
	void WriteSpecial (const SINX1_Platform::TChar *pSpecName,
		const SINX1_Platform::TChar *pData,int aHints=0)
	{
		DoWriteSpecial (pSpecName,pData,SINX1_Util::StrLen (pData),
			aHints|(*pData==SINX1_Platform::KCharNull? 0 : KHintTerminal));
	}

	/// @brief Version of WriteSpecial that takes data to write in form
	/// of SINX1_Util::String.
	void WriteSpecial (const SINX1_Platform::TChar *pSpecName,
		const SINX1_Util::String &aData,int aHints=0)
	{
		DoWriteSpecial (pSpecName,aData.c_str (),(unsigned)aData.size (),
			aHints|(aData.size ()>0? KHintTerminal : 0));
	}

	/// @brief Version of WriteSpecial that constructs data to write as string
	/// representation of integer aVal (signed decimal)
	template <class I,class U,int MAX_CHARS>
	void WriteSpecialCustomInt (const SINX1_Platform::TChar *pSpecName,
		I aVal,int aHints=0)
	{
		SINX1_Platform::TChar aString[MAX_CHARS],*pString = aString+MAX_CHARS-1;

		int i = 0;
		if (aVal>=0) DoPrepareCustomInt<U,MAX_CHARS>
			((U)aVal,false,false,pString,i);
		else DoPrepareCustomInt<U,MAX_CHARS> ((U)-aVal,false,true,pString,i);
		DoWriteSpecial (pSpecName,pString+1,(unsigned)i,aHints|KHintTerminal);
	}

	/// @brief Version of WriteSpecial that constructs data to write as string
	/// representation of integer aVal (signed decimal)
	void WriteSpecialInt (const SINX1_Platform::TChar *pSpecName,
		int aVal,int aHints=0)
	{
		WriteSpecialCustomInt<int,unsigned,64> (pSpecName,aVal,aHints);
	}

	/// @brief Version of WriteSpecial that constructs data to write as string
	/// representation of unsigned integer aVal (hexadecimal)
	template <class U,int MAX_CHARS>
	void WriteSpecialCustomUnsigned (const SINX1_Platform::TChar *pSpecName,
		unsigned aVal,int aHints=0)
	{
		SINX1_Platform::TChar aString[MAX_CHARS],*pString = aString+MAX_CHARS-1;

		int i = 0;
		DoPrepareInt (aVal,true,false,pString,i);
		DoWriteSpecial (pSpecName,pString+1,(unsigned)i,aHints|KHintTerminal);
	}

	/// @brief Version of WriteSpecial that constructs data to write as string
	/// representation of unsigned integer aVal (hexadecimal)
	void WriteSpecialUnsigned (const SINX1_Platform::TChar *pSpecName,
		unsigned aVal,int aHints=0)
	{
		WriteSpecialCustomUnsigned<unsigned,64> (pSpecName,aVal,aHints);
	}

	/// @brief Writes a comment type special char (<-- ...>) with given text
	/// as data, nicely formatted (<-- blah blah -->). (Quite a luxury already,
	/// don't you think so? :)
	/// All hints except for KHintNewline are ignored.
	void WriteSpecialComment (const SINX1_Util::String &aData,
		int aHints=0);

	/// @brief Return true if error has occured while writing the file. Error
	/// is not resettable. Any writing after error is set is silently ignored.
	bool CheckError () const;
};

/// @brief Class that incapsulates writing of complete SINX stream (one file,
/// that is). You are supposed to derive your own class from this one, define
/// interface functions that directly do with writing and add functions of your
/// own to open/finalize the actual stream. SINX_WriteDocumentBase is assumed
/// to operate on RAII principle (stream is opened (overwritten) as soon as you
/// create instance of your class and possibly call the open function, and is
/// closed as soon as instance is destroyed).
class SINX_WriteDocumentBase: public SINX_WriteNode
{
	friend class SINX_WriteNode;
	friend class SINX_WriteSpecialOpen;

	bool m_Error,m_Recursion;
	int m_IndentLevel;
	SINX_WriteNode *m_pLastOpen;

	void DoWriteData (const SINX1_Platform::TChar *pData,unsigned aCharCount)
	{
		if (!m_Error) m_Error |= !this->WriteData (pData,aCharCount);
	}

	void DoWriteCrLf ()
	{ if (!m_Error) m_Error |= !this->WriteCrLf (); }

protected:
	/// @brief Function that writes (appends) aCharCount chars to the stream,
	/// must return true if the write was successful of false if not.
	virtual bool WriteData
		(const SINX1_Platform::TChar *pData,unsigned aCharCount) = 0;

	/// @brief Function that writes (appends) end-of-line sequence to the
	/// stream, must return true if the write was successful of false if not.
	virtual bool WriteCrLf () = 0;

	/// @brief If your WriteData finds some TChar to be unwritable directly (due
	/// to encoding issues or whatever), it may call this function passing
	/// character code to it, and it will write the character as 'character
	/// code' special symbol using recursive call to WriteData. Don't call
	/// WriteCharAsCode for characters '<', '>', '#' and decimal digits, as
	/// they are themselves used to constitute the replacement sequence. If you
	/// try to, you will get an assertion failure.
	/// Returns the value that enclosed WriteData returned.
	bool WriteCharAsCode (int aCharCode)
	{
		SINX_ASSERT (!m_Recursion,
			"WriteCharAsCode: must not be called for #<>0-9");
		SINX_ASSERT (aCharCode>=0,
			"WriteCharAsCode: non-negative codes only are allowed");
		SINX1_Platform::TChar aString[64],*pString = aString+63;

		*(pString--) = SINX1_Util::CharFromCode ('>');
		int i = 1;
		DoPrepareInt ((unsigned)aCharCode,false,false,pString,i);
		*(pString--) = SINX1_Util::CharFromCode ('#');
		*(pString--) = SINX1_Util::CharFromCode ('<');
		m_Recursion = true;
		//bool aResult = WriteData (pString+1,i+3);
		bool aResult = WriteData (pString+1,i+2);
		m_Recursion = false;
		return aResult;
	}

	SINX_WriteDocumentBase ():
		m_Error (false),
		m_Recursion (false),
		m_IndentLevel (0)
	{
		m_pBase = this;
		m_pLastOpen = this;
		m_Compound = false;
		m_Empty = false; // to avoid auto-space in possible first WriteString
	}
};

class SINX_WriteSpecialOpen: public SINX_WriteNode
{
public:
	SINX_WriteSpecialOpen (SINX_WriteNode &aInsideOf,
		const SINX1_Platform::TChar *pSpecName,
		int aHints=SINX_WriteNode::KHintCompound)
	{
		SINX_ASSERT (aInsideOf.m_pBase!=NULL,
			"SINX_WriteSpecialOpen: use on a valid open SINX_WriteNode");
		PushOpen (&aInsideOf);
		SINX1_Platform::TChar
			aCharLAB = SINX1_Platform::KCharLAB,
			//aCharRAB = SINX1_Platform::KCharRAB,
			aCharSpace = SINX1_Util::CharFromCode (' '),
			aCharEqu = SINX1_Platform::KCharEqu,
			aCharDot = SINX1_Platform::KCharDot,
			aCharColon = SINX1_Platform::KCharColon;

		// write indent (if needed)
		if ((m_pPrev->m_Empty ||
			(m_pPrev->m_Compound && !m_pPrev->m_WasString)) &&
			!(aHints&SINX_WriteNode::KHintNewline))
			// write space if it's 1st character in the prev's data, or if
			// prev symbol is compound and other spec symbol has just been
			// written
			m_pBase->DoWriteData (&aCharSpace,1);
		if (aHints&SINX_WriteNode::KHintNewline)
		{
			m_pBase->DoWriteCrLf ();
			WriteIndent (m_pBase->m_IndentLevel-1);
			m_pPrev->m_NewlineUsed = true;
		}
		// write other things
		m_pBase->DoWriteData (&aCharLAB,1); // starting '<'
		if (aHints&KHintAttribute)
			// if attribute hint, prepend name with dot
			m_pBase->DoWriteData (&aCharDot,1);
		m_Empty = false;
		WriteString (pSpecName); // name
		if (aHints&KHintCompound)
		{
			// if compound hint, append colon to name
			m_pBase->DoWriteData (&aCharColon,1);
			m_Compound = true;
		}
		m_Empty = true;
		if (aHints&KHintTerminal)
		{
			// if terminal hint, write equ delimiter
			m_pBase->DoWriteData (&aCharEqu,1);
			m_Empty = false;
		}
	}

	~SINX_WriteSpecialOpen ()
	{
		SINX1_Platform::TChar
			aCharRAB = SINX1_Platform::KCharRAB;
		if (m_NewlineUsed)
		{
			// if there were spec subsymbols written with KHintNewline,
			// make a newline and indent before closing '>'
			m_pBase->DoWriteCrLf ();
			WriteIndent (m_pBase->m_IndentLevel-1);
		}
		m_pBase->DoWriteData (&aCharRAB,1); // closing '>'
		m_pPrev->m_Empty = false;
		m_pPrev->m_WasString = false;
	}

	/// @brief Shortcut to document writer's WriteCrLf
	void WriteCrLf () { m_pBase->WriteCrLf (); }
};

//
// implementations
//

inline bool SINX_WriteNode::CheckError () const
{
	return m_pBase==NULL || m_pBase->m_Error;
}

inline void SINX_WriteNode::PushOpen (SINX_WriteNode *pOver)
{
	m_pBase = pOver->m_pBase;
	m_pPrev = m_pBase->m_pLastOpen;
	m_pBase->m_pLastOpen = this;
	m_pBase->m_IndentLevel++;
}

inline void SINX_WriteNode::PopClose ()
{
	if (m_pBase!=NULL && m_pBase!=this)
	{
		m_pBase->m_IndentLevel--;
		// TODO: write closing '>', formatted according to hints
		m_pBase->m_pLastOpen = m_pPrev;
		m_pPrev->m_WasString = false;
		m_pBase = NULL;
	}
}

inline void SINX_WriteNode::WriteIndent (int aSpaceCount)
{
	if (m_pBase)
	{
		SINX1_Platform::TChar aSpaces[32];
		unsigned i,n = aSpaceCount/32,m = aSpaceCount%32;
		for (i=0; i<32; i++) aSpaces[i] = SINX1_Util::CharFromCode (' ');
		for (i=0; i<n; i++) m_pBase->DoWriteData (aSpaces,32);
		m_pBase->DoWriteData (aSpaces,m);
	}
}

inline void SINX_WriteNode::DoWriteSpecial (
	const SINX1_Platform::TChar *pSpecName,
	const SINX1_Platform::TChar *pData,unsigned aCharCount,
	int aHints)
{
	if (m_pBase)
	{
		// can only write to symbol's data if it's currently topmost open one
		SINX_ASSERT (m_pBase->m_pLastOpen==this,
			"WriteSpecial: can't write while there is unclosed inner symbol");

		SINX1_Platform::TChar
			aCharLAB = SINX1_Platform::KCharLAB,
			aCharRAB = SINX1_Platform::KCharRAB,
			aCharSpace = SINX1_Util::CharFromCode (' '),
			aCharEqu = SINX1_Platform::KCharEqu,
			aCharDot = SINX1_Platform::KCharDot,
			aCharColon = SINX1_Platform::KCharColon;
		// write indent (if needed)
		if ((m_pBase->m_pLastOpen->m_Empty ||
			(m_pBase->m_pLastOpen->m_Compound && !m_WasString))
			&& !(aHints&KHintNewline))
			// write space if it's 1st character in the data, or if
			// current symbol is compound and other spec symbol has just been
			// written
			m_pBase->DoWriteData (&aCharSpace,1);
		if (aHints&KHintNewline)
		{
			m_pBase->DoWriteCrLf ();
			WriteIndent (m_pBase->m_IndentLevel);
			m_NewlineUsed = true;
		}
		m_pBase->DoWriteData (&aCharLAB,1); // starting '<'
		if (aHints&KHintAttribute)
			// if attribute hint, prepend name with dot
			m_pBase->DoWriteData (&aCharDot,1);
		m_pBase->m_pLastOpen->m_Empty = false; // to skip space
		m_pBase->m_pLastOpen->WriteString (pSpecName); // name
		if (aHints&KHintCompound)
			// if compound hint, append colon to name
			m_pBase->DoWriteData (&aCharColon,1);
		if (aHints&KHintTerminal)
			// if terminal hint, write equ delimiter
			m_pBase->DoWriteData (&aCharEqu,1);
		else
			// otherwise, write space delimiter (if data is not empty)
			if (aCharCount>1) m_pBase->DoWriteData (&aCharSpace,1);
		// write the actual data
		m_pBase->m_pLastOpen->WriteString (pData,aCharCount);
		m_pBase->DoWriteData (&aCharRAB,1); // closing '>'
		// mark state changes to currently open symbol
		m_pBase->m_pLastOpen->m_Empty = false;
		m_pBase->m_pLastOpen->m_WasString = false;
	}
}

inline void SINX_WriteNode::WriteString (
	const SINX1_Platform::TChar *pData,unsigned aCharCount)
{
	if (m_pBase && !m_pBase->m_Error)
	{
		// can only write to symbol's data if it's currently topmost open one
		SINX_ASSERT (m_pBase->m_pLastOpen==this,
			"WriteString: can't write while there is unclosed inner symbol");

		if (aCharCount>0 && m_pBase->m_pLastOpen->m_Empty)
		{
			// if empty symbol, write space
			SINX1_Platform::TChar aCharSpace =
				SINX1_Util::CharFromCode (' ');
			m_pBase->DoWriteData (&aCharSpace,1);
		}

		if (aCharCount>0 && SINX1_Platform::IsWhitespace (*pData) &&
			m_pBase->m_pLastOpen->m_Empty)
		{
			// if string starts from whitespace and there was nothing written
			// to symbol's data yet, escape 1st character
			m_pBase->m_Error |=
				!(m_pBase->WriteCharAsCode (SINX1_Util::CodeFromChar (*pData)));
			if (m_pBase->m_Error) return;
			aCharCount--;
			pData++;
		}

		unsigned aLeft = aCharCount,aRunLen;
		SINX1_Platform::TChar aCharLAB = SINX1_Util::CharFromCode ('<'),
			aCharRAB = SINX1_Util::CharFromCode ('>'),
			aLT[4] = {
				aCharLAB,
				SINX1_Util::CharFromCode ('l'),
				SINX1_Util::CharFromCode ('t'),
				aCharRAB },
			aGT[4] = {
				aCharLAB,
				SINX1_Util::CharFromCode ('g'),
				SINX1_Util::CharFromCode ('t'),
				aCharRAB };
		for (;;)
		{
			if (aLeft==0) break;
			const SINX1_Platform::TChar *pRun = pData;
			for (aRunLen=0; aLeft>0 && *pData!=aCharLAB && *pData!=aCharRAB;
				aRunLen++, pData++, aLeft--) {}
			if (aRunLen>0)
				m_pBase->DoWriteData (pRun,aRunLen);
			if (aLeft>0)
			{
				m_pBase->DoWriteData (*pData==aCharLAB? aLT : aGT,4);
				pData++;
				aLeft--;
			}
		}

		m_pBase->m_pLastOpen->m_Empty &= aCharCount==0;
		m_pBase->m_pLastOpen->m_WasString = aCharCount>0;
	}
}

inline void SINX_WriteNode::DoWriteInt (unsigned aVal,bool aHex,bool aMinus)
{
	SINX1_Platform::TChar aString[64],*pString = aString+63;

	int i = 0;
	DoPrepareInt (aVal,aHex,aMinus,pString,i);
	m_pBase->m_Recursion = true;
	m_pBase->WriteData (pString+1,i);
	m_pBase->m_Recursion = false;
}

inline void SINX_WriteNode::WriteSpecialComment
	(const SINX1_Util::String &aData,int aHints)
{
	SINX1_Platform::TChar aOpening[3] =
	{
		SINX1_Util::KCharMinus,
		SINX1_Util::KCharMinus,
		SINX1_Platform::KCharNull
	},
	aClosing[3] =
	{
		SINX1_Util::CharFromCode (' '),
		SINX1_Util::KCharMinus,
		SINX1_Util::KCharMinus
	};
	SINX_WriteSpecialOpen rem (*this,aOpening,
		aHints&SINX_WriteNode::KHintNewline);
	rem.WriteString (aData);
	rem.WriteString (aClosing,3);
}

#endif // __SINX_WRITE_LEVEL1__
