//#include "sinx_read_l0.h"
//#include "sinx_read_l1.h" // includes one above
//#include "sinx_write.h" // includes both above
#include "sinx.h" // includes everything

// sample enum to show how to do with enums
enum TSampleEnum
{
	EUndef,
	EEmpty,
	EHidden
};

SINX::EnumHelper<TSampleEnum> SampleEnumHelper ()
{
	return SINX::EnumHelper<TSampleEnum> ()
		.Default ("undef", EUndef)
		("empty", EEmpty)
		("hidden", EHidden);
}

//
// writing sample
//

// (the purpose is to generate a complex SINX document, so don't get embarrassed
// by somewhat much code of writing example - there are many things written)

void SampleWriteAI (SINX::WriteNode &parent);
void SampleWriteLogic (SINX::WriteNode &parent);
void SampleWriteDiceLog (SINX::WriteNode &parent);

void SampleWrite ()
{
	printf ("Writing example_gen.sinx...\n");

	SINX::WriteFileOpen outf ("example_gen.sinx");
	if (!outf.IsOpenOk ())
	{
		printf ("Failed to overwrite example_gen.sinx\n");
		exit (1);
	}

	// examples of writing strings (terminal chars)
	outf.WriteString ("Check of writing some text");
	outf.WriteCrLf ();
	outf.WriteString ("Check of writing something else (a text too)");
	outf.WriteCrLf ();
	outf.WriteString ("Strings <are escaped>");

	// examples of writing special chars with directly specified data
	// (such chars are written with terminal attr if the data is not empty)
	outf.WriteCrLf ();
	outf.WriteString ("Example of empty spec symbol: ");
	outf.WriteSpecial ("test","");
	outf.WriteSpecial ("test","0 (has newline formatter hint)",
		SINX::KHintNewline);
	outf.WriteSpecial ("test","1 (attribute hint)",
		SINX::KHintAttribute);
	outf.WriteSpecial ("test"," 2 (starts with whitespace, has both attr and newline hints)",
		SINX::KHintNewline|SINX::KHintAttribute);

	// examples of writing special chars with compound data
	// something compound
	outf.WriteCrLf ();
	outf.WriteCrLf ();
	outf.WriteString ("Example of something compound:");
	{
		// create <config: ...> node
		SINX::WriteSpecialOpen config (outf,"config",
			SINX::KHintNewline|SINX::KHintCompound);

		// a comment for better readability
		outf.WriteSpecialComment
			("program configuration, really",SINX::WriteNode::KHintNewline);

		// create <.sfx=on> and some other keys in config
		// tell writer to place each one at own line
		SINX::WriteSpecialBoolL (config,"sfx",true,
			SINX::KHintNewline|SINX::KHintAttribute);
		SINX::WriteSpecialStringL (config,"soundtrack","epic",
			SINX::KHintNewline|SINX::KHintAttribute);
		SINX::WriteSpecialBoolL (config,"tips",false,
			SINX::KHintNewline|SINX::KHintAttribute);
		// example of entry with numeric data (SINX data is string only, but
		// lib provides int/unsigned reading/writing helpers for convenience)
		SINX::WriteSpecialUnsignedL (config,"mask",65533,
			SINX::KHintNewline|SINX::KHintAttribute);

		// create some entries iteratively and write data from string
		std::basic_string<SINX::TChar> aExample;
		int i;
		for (i=0; i<4; i++)
		{
			aExample += "g";
			config.WriteSpecial ("test",aExample,SINX::KHintNewline);
		}
	}

	// something heavily compound
	outf.WriteCrLf ();
	outf.WriteCrLf ();
	outf.WriteString ("Example of something heavily compound:");
	{
		// create <data: ...> node
		SINX::WriteSpecialOpen data (outf,"data",
			SINX::KHintNewline|SINX::KHintCompound);

		{
			// create <meta: ...> node under data
			SINX::WriteSpecialOpen meta (data,"meta",
				SINX::KHintNewline|SINX::KHintCompound);

			// place some entries under it, all in one line
			meta.WriteSpecial ("version","1.0",
				SINX::WriteNode::KHintAttribute);
			meta.WriteSpecial ("blahblah","blah blah blah",
				SINX::WriteNode::KHintAttribute);
		}

		// write some more compound subnodes
		SampleWriteAI (data);
		SampleWriteLogic (data);
		SampleWriteDiceLog (data);
	}
}

void SampleWriteAI (SINX::WriteNode &parent)
{
	// create <ai: ...> node
	SINX::WriteSpecialOpen logic (parent,"ai",
		SINX::KHintNewline|SINX::KHintCompound);

	// create some entries
	int i,mtgts[] = { 3, 256, 7, 11, 22, -1 };
	for (i=0; mtgts[i]!=-1; i++)
	{
		SINX::WriteSpecialOpen entry (parent,"entry",
			SINX::KHintNewline|SINX::KHintCompound);

		SINX::WriteSpecialIntL (entry,"id",i,0);
		{
			// create <data: ...> subentry
			SINX::WriteSpecialOpen data (entry,"data",
				SINX::KHintCompound);

			data.WriteSpecial ("atgt","rnd");
			SINX::WriteSpecialIntL (data,"mtgt",mtgts[i],0);
		}
	}
}

void SampleWriteLogic (SINX::WriteNode &parent)
{
	// create <logic: ...> node
	SINX::WriteSpecialOpen logic (parent,"logic",
		SINX::KHintNewline|SINX::KHintCompound);

	{
		// create <meta: ...> node under logic
		SINX::WriteSpecialOpen meta (logic,"meta",
			SINX::KHintNewline|SINX::KHintCompound);

		// place some entries under it, all in one line
		SINX::WriteSpecialStringL (meta,"api_version","1.0",
			SINX::KHintAttribute);
	}

	{
		// create <map: ...> node under logic
		SINX::WriteSpecialOpen meta (logic,"map",
			SINX::KHintNewline|SINX::KHintCompound);

		// place width and height under it in one line
		SINX::WriteSpecialIntL (meta,"width",32,
			SINX::KHintNewline|SINX::KHintAttribute);
		SINX::WriteSpecialIntL (meta,"height",16,
			SINX::KHintAttribute);

		// place some <cell: ...> nodes under map, each on its own line
		int i;
		SINX::EnumHelper<TSampleEnum> aEnumHlp = SampleEnumHelper ();
		for (i=0; i<8; i++)
		{
			SINX::WriteSpecialOpen cell (logic,"cell",
				SINX::KHintNewline|SINX::KHintCompound);

			SINX::WriteSpecialEnumL (cell,"type",aEnumHlp((i&1)? EEmpty : EHidden),
				SINX::KHintAttribute);
			SINX::WriteSpecialIntL (cell,"charges",i,
				SINX::KHintAttribute);
		}
	}
}

void SampleWriteDiceLog (SINX::WriteNode &parent)
{
	// create <dice_log: ...> node
	SINX::WriteSpecialOpen dice_log (parent,"dice_log",
		SINX::KHintNewline|SINX::KHintCompound);

	// create some entries
	int i,ds[] = { 3, 0, 12, 0, -1 };
	for (i=0; ds[i]!=-1; i++)
	{
		SINX::WriteSpecialOpen dice (parent,"dice",
			SINX::KHintNewline|SINX::KHintCompound);

		if (ds[i]==0)
			dice.WriteSpecial ("d","rnd",SINX::KHintAttribute);
		else
			dice.WriteSpecialInt ("d",ds[i],SINX::KHintAttribute);

		dice.WriteSpecialInt ("n",1,SINX::KHintAttribute);
		dice.WriteSpecial ("p","EMonsterGen",SINX::KHintAttribute);
	}
}

//
// reading sample
//

void SampleRead ()
{
	printf ("Reading example_gen.sinx...\n");

	SINX::ReadFileOpen inf ("example_gen.sinx");
	if (!inf.IsOpenOk ())
	{
		printf ("Failed to read example_gen.sinx\n");
		exit (1);
	}

	// check map (access via OpenPathL)
	printf ("Checking map...\n");
	SINX::Symbol smap = inf.OpenPathL ("<data <logic <map>>>");
	// (if /data/logic/map doesn't exist, smap will be an empty symbol)
	printf ("Map width: %d\n",smap["width"].DataAsIntL (-1));
	// (if smap is an empty symbol or has no width subsymbol, smap["width"]
	// is an empty symbol, and DataAsIntL returns default, which is -1)
	printf ("Map height: %d\n",smap["height"].DataAsIntL (-1));
	// (same about height)
	SINX::Symbol cell6 = inf.OpenPathL ("<data <logic <map <cell><.index=6>>>>");
	// or so: smap.OpenPathL ("<cell><.index=6>");
	printf ("Map cell #6 type: %s\n",cell6["type"].DataAsStringL ().c_str ());
	SINX::EnumHelper<TSampleEnum> aEnumHlp = SampleEnumHelper ();
	printf ("Map cell #6 internal type is: %d\n",
		int (ReadSpecialEnumL (cell6["type"],aEnumHlp)));

	// check config (dictionary access method)
	printf ("Checking config...\n");
	SINX::Symbol config = inf["config"];
	printf ("Mask (default is 0): %d\n",config["mask"].DataAsIntL (0));
	std::deque<SINX::Symbol> tests = config.GetAllSpecialL ("test");
	int i;
	printf ("Test entries in config: %d\n",tests.size ());
	for (i=0; i<(int)tests.size (); i++)
		printf ("Test %d value: %s\n",i,tests[i].DataAsStringL ().c_str ());
	printf ("Value of last test entry in config: %s\n",
		// addressing via dictionary method to name that is shared by several
		// symbols will yield last of them
		inf["config"]["test"].DataAsStringL ().c_str ());
}

//
// reading sample 2 - dump
//

void DumpSymbol (const SINX::Symbol &aSym,int aIndent)
{
	std::deque<SINX::Symbol::Fragment> aFrags = aSym.GetAllL ();
	int i;
	for (i=0; i<(int)aFrags.size (); i++)
	if (aFrags[i].IsString ())
		printf ("%*s- STRING: %s\n",aIndent,"",aFrags[i].AsStringL ().c_str ());
	else
	{
		printf ("%*s- SPECIAL: %s\n",aIndent,"",
			aFrags[i].AsSpecial ().NameL ().c_str ());
		DumpSymbol (aFrags[i].AsSpecial (),aIndent+1);
	}
}

void SampleDump (char *pName)
{
	printf ("Reading %s...\n",pName);

	SINX::ReadFileOpen inf (pName);
	if (!inf.IsOpenOk ())
	{
		printf ("Failed to read %s\n",pName);
		exit (1);
	}
	DumpSymbol (inf,0);
}

int main (int argc,char *argv[])
{
	try {
	if (argc<2)
	{
		SampleWrite ();
		SampleRead ();
		printf ("Now try 'test.exe filename.sinx' to parse and dump a random SINX file!\n");
	}
	else
		SampleDump (argv[1]);
	}
	catch (...)
	{
	}

	return 0;
}