#ifndef __SINX_H__
#define __SINX_H__

#include "sinx_write.h" // also includes sinx_read_l0-1.h

//
// in order to have a consistent writer, we must define a document writer class
// inherited from SINX_WriteDocumentBase that works with particular output
//

/// @brief SINX format file writer. Use: SINX_WriteFileOpen outf ("filename"),
/// check by IsOpenOk, then exploit outf with methods.
class SINX_WriteFileOpen: public SINX_WriteDocumentBase
{
	std::string m_Filename;
	FILE *m_File;
	bool m_OpenOk;

public:
	SINX_WriteFileOpen (const std::string &aFilename);
	~SINX_WriteFileOpen ();
	bool IsOpenOk () const { return m_OpenOk; }

	//
	// implementations of SINX_WriteDocumentBase required interface
	//

	bool WriteData (const SINX1_Platform::TChar *pData,unsigned aCharCount);
	bool WriteCrLf ();
};

/// @brief SINX format file reader. Use: SINX_ReadFileOpen inf ("filename"),
/// check by IsOpenOk (not actually needed, failure will leave this as an empty
/// symbol), then exploit inf with methods (as SINX1_Symbol).
class SINX_ReadFileOpen: public SINX1_Symbol
{
	bool m_OpenOk;
	FILE *m_File;

public:
	SINX_ReadFileOpen (const std::string &aFilename);
	~SINX_ReadFileOpen ();
	bool IsOpenOk () const { return m_OpenOk; }
};

//
// define some shortcuts for convenience
//
namespace SINX
{
	typedef SINX0_Platform Platform;
	typedef Platform::TChar TChar; // it's char here
	typedef std::basic_string<Platform::TChar> StdString;

	typedef SINX1_Symbol Symbol;

	typedef SINX_WriteNode WriteNode;
	typedef SINX_WriteSpecialOpen WriteSpecialOpen;
	typedef SINX_WriteDocumentBase WriteDocumentBase;

	typedef SINX_WriteFileOpen WriteFileOpen;
	typedef SINX_ReadFileOpen ReadFileOpen;

	StdString ReadSpecialStringL (const Symbol &,const StdString &aDefault);
	void WriteSpecialStringL (WriteNode &,const TChar *pNewSpecName,
		const StdString &aData,int aHints);
	template <class I>
	I ReadSpecialCustomIntL (const Symbol &aFrom,I aDefault)
	{
		return aFrom.IsEof ()? aDefault : aFrom.DataAsCustomIntL<I> (aDefault);
	}
	template <class I,class U,int MAX_CHARS>
	void WriteSpecialCustomIntL (WriteNode &aTo,const TChar *pNewSpecName,
		I aData,int aHints)
	{
		aTo.WriteSpecialCustomInt<I,U,MAX_CHARS> (pNewSpecName,aData,aHints);
	}
	int ReadSpecialIntL (const Symbol &,int aDefault);
	void WriteSpecialIntL (WriteNode &,const TChar *pNewSpecName,
		int aData,int aHints);
	int ReadSpecialUnsignedL (const Symbol &,unsigned aDefault);
	void WriteSpecialUnsignedL (WriteNode &,const TChar *pNewSpecName,
		unsigned aData,int aHints);
	bool ReadSpecialBoolL (const Symbol &,bool aDefault);
	void WriteSpecialBoolL (WriteNode &,const TChar *pNewSpecName,
		bool aData,int aHints);

	template <class Enum>
	class EnumHelper;
	template <class Enum>
	Enum ReadSpecialEnumL (const Symbol &aFrom,const EnumHelper<Enum> &aEH);
	template <class Enum>
	void WriteSpecialEnumL (WriteNode &aTo,const TChar *pNewSpecName,
		const EnumHelper<Enum> &aEH,int aHints);

	/// @brief Helper to serialize/deserialize enums. Template paramer type is
	/// the enum type (or may be int/unsigned/etc).
	/// Use: 1) prepare (EnumHelper<YourEnumType> eh;
	/// eh ("name1", value1) ("name2", value2) ...
	/// .Default ("name_for_unrecognized_values", value_for_unrecognized_names);
	/// 2) read/write via ReadSpecialEnumL (const Symbol &,eh),
	/// WriteSpecialEnumL (const Symbol &,"nodename",eh(value)),
	template <class Enum>
	class EnumHelper
	{
		friend Enum ReadSpecialEnumL<Enum>
			(const Symbol &aFrom,const EnumHelper<Enum> &aEH);
		friend void WriteSpecialEnumL<Enum> (WriteNode &aTo,
			const TChar *pNewSpecName,const EnumHelper<Enum> &aEH,int aHints);

		std::map<StdString,Enum> m_ReadMap;
		std::map<Enum,StdString> m_WriteMap;

		StdString m_DefaultName;
		Enum m_DefaultValue,m_ValueToWrite;

	public:
		EnumHelper &operator () (const StdString &aName,Enum aValue)
		{
			m_ReadMap[aName] = aValue;
			m_WriteMap[aValue] = aName;
			return *this;
		}

		EnumHelper &Default (const StdString &aName,Enum aValue)
		{
			m_DefaultName = aName;
			m_DefaultValue = aValue;
			return *this;
		}

		/// @brief Marks intent to write enum value aValue using this helper.
		/// Intended usage is with WriteSpecialEnumL<Enum>:
		/// WriteSpecialEnumL (aWriteNode,"nodename",
		/// aYourEnumHelper (value_to_write))
		EnumHelper &operator () (Enum aValue)
		{
			m_ValueToWrite = aValue;
			return *this;
		}
	};

	template <class Enum>
	Enum ReadSpecialEnumL (const Symbol &aFrom,const EnumHelper<Enum> &aEH)
	{
		typename std::map<StdString,Enum>::const_iterator it =
			aEH.m_ReadMap.find (aFrom.DataAsStringL ());
		if (it!=aEH.m_ReadMap.end ())
			return it->second;
		else
			return aEH.m_DefaultValue;
	}
	/// @brief Assumed to use via
	/// WriteSpecialEnumL (aSomeWriteNode,"nodename",aSomeEH(value_to_write))
	template <class Enum>
	void WriteSpecialEnumL (WriteNode &aTo,const TChar *pNewSpecName,
		const EnumHelper<Enum> &aEH,int aHints)
	{
		typename std::map<Enum,StdString>::const_iterator it =
			aEH.m_WriteMap.find (aEH.m_ValueToWrite);
		if (it!=aEH.m_WriteMap.end ())
			WriteSpecialStringL (aTo,pNewSpecName,it->second,aHints);
		else
			WriteSpecialStringL (aTo,pNewSpecName,aEH.m_DefaultName,aHints);
	}

	/// @brief Flag for WriteSpecial, denotes that symbol should be emitted
	/// with attribute hint
	static const int KHintAttribute = WriteNode::KHintAttribute;

	/// @brief Flag for WriteSpecial, denotes that symbol should be emitted
	/// with compound hint
	static const int KHintCompound = WriteNode::KHintCompound;

	/// @brief Flag for WriteSpecial, denotes that symbol should be emitted
	/// with terminal hint
	static const int KHintTerminal = WriteNode::KHintTerminal;

	/// @brief Flag for WriteSpecial, denotes that symbol should be emitted
	/// with newline and indent. Use inside symbols opened with KHintCompound
	/// to arrange a pretty formatting.
	static const int KHintNewline = WriteNode::KHintNewline;
};

#endif // __SINX_H__
